@@ -31,9 +31,11 @@
 #include "http_request.h"
 #include "apr_version.h"
 #include "apr_ldap.h"
-#include "apr_strings.h"
 #include "apr_reslist.h"
+#include "apr_strings.h"
+#include "apr_tables.h"
 #include "util_ldap.h"
+#include "util_script.h"
 
 #if !defined(APU_HAS_LDAP) && !defined(APR_HAS_LDAP)
 #error mod_vhost_ldap requires APR-util to have LDAP support built in
@@ -50,6 +52,8 @@
 #define MIN_UID 100
 #define MIN_GID 100
 
+#define MAX_FAILURES 5
+
 module AP_MODULE_DECLARE_DATA vhost_ldap_module;
 
 typedef enum {
@@ -115,6 +119,38 @@ static void ImportULDAPOptFn(void)
 }
 #endif 
 
+/* Taken from server/core.c */
+static int set_document_root(request_rec *r, const char *arg)
+{
+    void *sconf = r->server->module_config;
+    core_server_config *conf = ap_get_module_config(sconf, &core_module);
+
+    /* Make it absolute, relative to ServerRoot */
+    arg = ap_server_root_relative(r->pool, arg);
+
+    if (arg == NULL) {
+        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r, 
+                      "[mod_vhost_ldap.c] set_document_root: DocumentRoot [%s] must be a directory",
+		      arg);
+
+        return HTTP_INTERNAL_SERVER_ERROR;
+    }
+
+    /* TODO: ap_configtestonly && ap_docrootcheck && */
+    if (apr_filepath_merge((char**)&conf->ap_document_root, NULL, arg,
+                           APR_FILEPATH_TRUENAME, r->pool) != APR_SUCCESS
+        || !ap_is_directory(r->pool, arg)) {
+
+        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0,
+		      r,
+		      "[mod_vhost_ldap.c] set_document_root: Warning: DocumentRoot [%s] does not exist",
+		      arg);
+        conf->ap_document_root = arg;
+    }
+    return OK;
+}
+
+
 static int mod_vhost_ldap_post_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
 {
     /* make sure that mod_ldap (util_ldap) is loaded */
@@ -423,20 +459,21 @@ static int mod_vhost_ldap_translate_name(request_rec *r)
 {
     request_rec *top = (r->main)?r->main:r;
     mod_vhost_ldap_request_t *reqc;
-    apr_table_t *e;
     int failures = 0;
     const char **vals = NULL;
     char filtbuf[FILTER_LENGTH];
     mod_vhost_ldap_config_t *conf =
 	(mod_vhost_ldap_config_t *)ap_get_module_config(r->server->module_config, &vhost_ldap_module);
-    core_server_config * core =
-	(core_server_config *) ap_get_module_config(r->server->module_config, &core_module);
     util_ldap_connection_t *ldc = NULL;
     int result = 0;
     const char *dn = NULL;
     char *cgi;
     const char *hostname = NULL;
     int is_fallback = 0;
+    int sleep0 = 0;
+    int sleep1 = 1;
+    int sleep;
+    struct berval hostnamebv, shostnamebv;
 
     reqc =
 	(mod_vhost_ldap_request_t *)apr_pcalloc(r->pool, sizeof(mod_vhost_ldap_request_t));
@@ -459,17 +496,24 @@ start_over:
     else {
         ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, r, 
                       "[mod_vhost_ldap.c] translate: no conf->host - weird...?");
-        return DECLINED;
+        return HTTP_INTERNAL_SERVER_ERROR;
     }
 
     hostname = r->hostname;
+    if (hostname == NULL || hostname[0] == '\0')
+        goto null;
 
 fallback:
 
     ap_log_rerror(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r,
-		   "[mod_vhost_ldap.c]: translating %s", r->uri);
+		  "[mod_vhost_ldap.c]: translating hostname [%s], uri [%s]",
+		  hostname, r->uri);
 
-    apr_snprintf(filtbuf, FILTER_LENGTH, "(&(%s)(|(apacheServerName=%s)(apacheServerAlias=%s)))", conf->filter, hostname, hostname);
+    ber_str2bv(hostname, 0, 0, &hostnamebv);
+    if (ldap_bv2escaped_filter_value(&hostnamebv, &shostnamebv) != 0)
+	goto null;
+    apr_snprintf(filtbuf, FILTER_LENGTH, "(&(%s)(|(apacheServerName=%s)(apacheServerAlias=%s)))", conf->filter, shostnamebv.bv_val, shostnamebv.bv_val);
+    ber_memfree(shostnamebv.bv_val);
 
     result = util_ldap_cache_getuserdn(r, ldc, conf->url, conf->basedn, conf->scope,
 				       attributes, filtbuf, &dn, &vals);
@@ -477,13 +521,38 @@ fallback:
     util_ldap_connection_close(ldc);
 
     /* sanity check - if server is down, retry it up to 5 times */
-    if (result == LDAP_SERVER_DOWN) {
-        if (failures++ <= 5) {
+    if (AP_LDAP_IS_SERVER_DOWN(result) ||
+	(result == LDAP_TIMEOUT) ||
+	(result == LDAP_CONNECT_ERROR)) {
+        sleep = sleep0 + sleep1;
+        ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, r,
+		      "[mod_vhost_ldap.c]: lookup failure, retry number #[%d], sleeping for [%d] seconds",
+		      failures, sleep);
+        if (failures++ < MAX_FAILURES) {
+	    /* Back-off exponentially */
+	    apr_sleep(apr_time_from_sec(sleep));
+	    sleep0 = sleep1;
+	    sleep1 = sleep;
             goto start_over;
-        }
+        } else {
+	    return HTTP_GATEWAY_TIME_OUT;
+	}
     }
 
-    if ((result == LDAP_NO_SUCH_OBJECT)) {
+    if (result == LDAP_NO_SUCH_OBJECT) {
+	if (strcmp(hostname, "*") != 0) {
+	    if (strncmp(hostname, "*.", 2) == 0)
+		hostname += 2;
+	    hostname += strcspn(hostname, ".");
+	    hostname = apr_pstrcat(r->pool, "*", hostname, NULL);
+	    ap_log_rerror(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, 0, r,
+		          "[mod_vhost_ldap.c] translate: "
+			  "virtual host not found, trying wildcard %s",
+			  hostname);
+	    goto fallback;
+	}
+
+null:
 	if (conf->fallback && (is_fallback++ <= 0)) {
 	    ap_log_rerror(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, 0, r,
 			  "[mod_vhost_ldap.c] translate: "
@@ -498,7 +567,7 @@ fallback:
 		      "virtual host %s not found",
 		      hostname);
 
-	return DECLINED;
+	return HTTP_BAD_REQUEST;
     }
 
     /* handle bind failure */
@@ -507,7 +576,7 @@ fallback:
                       "[mod_vhost_ldap.c] translate: "
                       "translate failed; virtual host %s; URI %s [%s]",
 		      hostname, r->uri, ldap_err2string(result));
-	return DECLINED;
+	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
     /* mark the user and DN */
@@ -554,7 +623,7 @@ fallback:
         ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r, 
                       "[mod_vhost_ldap.c] translate: "
                       "translate failed; ServerName or DocumentRoot not defined");
-	return DECLINED;
+	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
     cgi = NULL;
@@ -566,12 +635,23 @@ fallback:
 	}
     }
     if (cgi) {
-	r->filename = apr_pstrcat (r->pool, reqc->cgiroot, cgi + strlen("cgi-bin"), NULL);
-	r->handler = "cgi-script";
-	apr_table_setn(r->notes, "alias-forced-type", r->handler);
+        /* Set exact filename for CGI script */
+        cgi = apr_pstrcat(r->pool, reqc->cgiroot, cgi + strlen("cgi-bin"), NULL);
+        if ((cgi = ap_server_root_relative(r->pool, cgi))) {
+	  ap_log_rerror(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r,
+			"[mod_vhost_ldap.c]: ap_document_root is: %s",
+			ap_document_root(r));
+	  r->filename = cgi;
+	  r->handler = "cgi-script";
+	  apr_table_setn(r->notes, "alias-forced-type", r->handler);
+	}
     } else if (r->uri[0] == '/') {
-	r->filename = apr_pstrcat (r->pool, reqc->docroot, r->uri, NULL);
+        /* we don't set r->filename here, and let other modules do it
+         * this allows other modules (mod_rewrite.c) to work as usual
+	 */
+        /* r->filename = apr_pstrcat (r->pool, reqc->docroot, r->uri, NULL); */
     } else {
+        /* We don't handle non-file requests here */
 	return DECLINED;
     }
 
@@ -581,16 +661,14 @@ fallback:
 	top->server->server_admin = apr_pstrdup (top->pool, reqc->admin);
     }
 
-    // set environment variables
-    e = top->subprocess_env;
-    apr_table_addn (e, "SERVER_ROOT", reqc->docroot);
+    result = set_document_root(r, reqc->docroot);
+    if (result != OK) {
+        return HTTP_INTERNAL_SERVER_ERROR;
+    }
+    apr_table_setn(r->notes, "vhost-document-root", reqc->docroot);
 
-    core->ap_document_root = apr_pstrdup(top->pool, reqc->docroot);
-
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r,
-		  "[mod_vhost_ldap.c]: translated to %s", r->filename);
-
-    return OK;
+    /* Hack to allow post-processing by other modules (mod_rewrite, mod_alias) */
+    return DECLINED;
 }
 
 #ifdef HAVE_UNIX_SUEXEC
@@ -635,17 +713,36 @@ static ap_unix_identity_t *mod_vhost_ldap_get_suexec_id_doer(const request_rec *
 }
 #endif
 
+static int mod_vhost_ldap_fixups(request_rec *r)
+{
+    char *docroot = 
+      apr_table_get(r->notes, "vhost-document-root");
+
+    if (docroot != NULL)
+      return set_document_root(r, );
+    else
+      return DECLINED;
+}
+
 static void
 mod_vhost_ldap_register_hooks (apr_pool_t * p)
 {
+
+    /*
+     * Run before mod_rewrite
+     */
+    static const char * const aszRewrite[]={ "mod_rewrite.c", NULL };
+
     ap_hook_post_config(mod_vhost_ldap_post_config, NULL, NULL, APR_HOOK_MIDDLE);
-    ap_hook_translate_name(mod_vhost_ldap_translate_name, NULL, NULL, APR_HOOK_MIDDLE);
+    ap_hook_translate_name(mod_vhost_ldap_translate_name, NULL, aszRewrite, APR_HOOK_FIRST);
 #ifdef HAVE_UNIX_SUEXEC
     ap_hook_get_suexec_identity(mod_vhost_ldap_get_suexec_id_doer, NULL, NULL, APR_HOOK_MIDDLE);
 #endif
 #if (APR_MAJOR_VERSION >= 1)
     ap_hook_optional_fn_retrieve(ImportULDAPOptFn,NULL,NULL,APR_HOOK_MIDDLE);
 #endif
+
+    ap_hook_fixups(mod_vhost_ldap_fixups, NULL, NULL, APR_HOOK_LAST);
 }
 
 module AP_MODULE_DECLARE_DATA vhost_ldap_module = {
